#include <cassert>
#include <filesystem>
#include <fstream>
#include <iostream>
#include <print>
#include <sstream>
#include <utility>

// 0. Ülesanne
// Kirjuta main-funktsioon, signatuuriga `int main(int argc, char **argv)` ning
// konstrueeri vastavad konteinerid argumentide jaoks, et saaks välja kutsuda
// funktsiooni `MinuMain` (vt signatuuri all). NB! Funktsiooni `MinuMain`
// signatuuri ära muuda! Selle olemasolu ja korrektsust kontrollib rida
//   `static_assert(MinuMainKontroll<>);`
// Mida see rida täpselt teeb, vaatame järgnevatel nädalatel.
//
// Funktsiooni `main` mall on täiesti faili alguses. Ära seda mujale liiguta,
// oma lahendus kirjuta sinna sisse.

// Järgnev rida lubab meil `std::filesystem` asemel kasutada `fs`, mis teeb
// koodi mugavalt lihtsamaks.
namespace fs = std::filesystem;

// Kodutöö ZIP-arhiivis on kaasa pakitud kaust `failid`. Vaata, et see oleks su
// projektis kätte saadav.
//
// Iga programmi käivitamisel pannakse see käima mingist kindlast kaustast. Seda
// kausta kutsutakse "working directory". Uuri lähemalt:
// https://en.wikipedia.org/wiki/Working_directory.
//
// CLionis programmi jooksutamisel on working directory reeglina
// `cmake-build-debug`, mille olemasolu ja haldamisega tegeleb tööriist CMake.
//
// Kui tahame failiteed viidata ZIP-arhiivis olevale kaustale `failid`, siis
// peame minema ühe kausta võrra tagasi (`..`) ja siis saame ligipääsu antud
// kaustale.
//
// Failipuu näide:
//   cmake-build-debug/
//     ... <- siin kaustas on kuskil programm ise (reeglina nimega, mis on
//     määratud CMakeLists.txt failis)
//            seega programmi käivitamisel on working directory just see kaust
//            `cmake-build-debug/`
//   failid/
//     ...
//   main.cpp <- see fail
//   CMakeLists.txt
//
// Kui jooksutad programmi kuidagi teisiti, muuda allolevat muutujat vastavalt
// nii, et see viitaks antud kaustale.
const auto FailidDir = fs::path("../failid");

// 1. Ülesanne
//
// Kirjuta funktsioon `readFileUntil(fs::path, long n)`, mis loeb faili ja
// tagastab selle sisu vastavalt järgmistele reeglitele:
//   1. Kui tegemist on tekstifailiga (laiendiga .txt, .sql, .cpp jne), siis
//   tagastab funktsioon ülimalt `n` karakterit faili algusest.
//     Kui faili sisu on lühem, siis on sõne pikkus sama, mis karakterite arv
//     failis.
//   2. Kui tegemist on binaarfailiga (fail, millel pole faililaiendit), siis
//   loeb failist sisu bait-haaval ning lõpetab
//     töö, kui leiab funktsioonist väärtuse 0. Funktsioon tagastab baitide
//     numbrilised väärtused sõnena. See tähendab, et kui failis on baidid `4 15
//     0`, tagastab funktsioon `4150`. ZIP-arhiivis kaasa antud failide kohta
//     leiad testid alt.
// Võid eeldada, et funktsioon kutsutakse välja ainult failidel, s.t kaustade
// korral ei pea funktsioon midagi tegema. Funktsioon peab samuti kontrollima
// faili avamisel selle olemasolu. Kui faili ei leidu, tagastab funktsioon tühja
// sõne.
std::string readFileUntil(const fs::path &path, const long n) {
  std::ifstream f(path);
  assert(f);

  std::vector<char> buf(n);
  f.read(buf.data(), n);

  auto fPos = path.string().rfind('/');
  auto fileName = path.string().substr(fPos);
  if (fileName.find('.') == std::string::npos) {
    // Eelduste järgi binaarfail
    int last = -1;
    std::stringstream ss;
    for (auto c : buf) {
      if (last == 0 && c == 0) {
        break;
      }

      last = static_cast<int>(static_cast<unsigned char>(c));
      ss << std::hex << last;
    }
    return ss.str();
  }

  return {buf.data()};
}

// Defineeri üks tüübinimi (alias, vihje: `using`), mis vastaks sõnepaarile.
// Selle kohta uuri klassi `std::pair<T, U>`:
// https://en.cppreference.com/w/cpp/utility/pair. `StrPair paar{"a", "b"};`
// peaks olema valiidne C++ kood.
using StrPair = std::pair<std::string, std::string>;

// Allolev kood kontrollib, kas määrasid tüübialiase õigesti
template <typename T = std::string>
concept StrPairKontroll = requires(T a, T b) {
  { StrPair{a, b} };
};
static_assert(StrPairKontroll<>);

// 2. Ülesanne
// Kirjuta funktsioon `readPairs(fs::path path)`, mis loeb ja töötleb antud
// faili rea kaupa. Iga rida, mis on kujul "<sõna> ja <sõna>" on valiidne rida,
// tuleb konstrueerida sõnapaariks (vt üleval defineeritud tüübialiast) ning
// lisada vektorisse. Faili sisu lõppedes tagastab funktsioon paaride vektorid.
// Funktsioon võib eeldada, et fail eksisteerib ning antud asukoht ei viita
// kaustale. Funktsioon võib eeldada, et failis on vähemalt üks rida, mis EI
// pruugi olla õiges formaadis.
//
// Lisa: Uuri lähemalt klassi `std::vector` liikmefunktsiooni `emplace_back`.
// Selle kasutamine ei ole kohustuslik, aga äkki on selles ülesandes
// 'efektiivsem'?.
std::vector<StrPair> readPairs(const fs::path &path) {
  std::ifstream f(path);
  assert(f);

  std::vector<StrPair> pairs{};
  std::string rida;
  while (std::getline(f, rida)) {
    auto const jaIt = rida.find(" ja ");
    if (jaIt == std::string::npos) {
      continue;
    }

    const auto start = rida.substr(0, jaIt);
    const auto end = rida.substr(jaIt + 4);
    if (!start.empty() && !end.empty()) {
      pairs.emplace_back(start, end);
    }
  }

  return pairs;
}

// 3. Ülesanne
// Kirjuta funktsioon `walk(fs::path)`, mis käitub järgmiselt:
//   Kui tegemist on kausta asukohaga (s.t path viitab kaustale), siis väljastab
//   kausta asukoha lõpuga "/" ning
//     käitub kausta alamfailide ja -kaustade korral täpselt samamoodi nagu
//     kirjeldatud (vihje: rekursioon). Pane tähele, et järjekord ei ole
//     oluline.
//   Kui tegemist on faili asukohaga, siis väljastab faili 10 esimest karakterit
//   (kasuta funktsiooni `readFileUntil`.
//
// Näide: Vaata kodutöö ZIP-arhiivis kaasa antud kausta `failid`.
// Selle funktsiooni välja kutsumisel argumendiga `FAILID_DIR` (vt makro
// kommentaare!) peaks olema tulemus umbes järgnev:
// ../failid/
// ../failid/b.sql -> 'SELECT * F'
// ../failid/a.txt -> 'see on fai'
// ../failid/c/
// ../failid/c/d.cpp -> '#include <'
// ../failid/c/f.txt -> 'lyhike'
// ../failid/c/e -> '69420'
// ../failid/paarid.txt -> 'must ja va'
//
// NB! Pane tähele, et olenevalt sellest, kas jooksutad programmi läbi CLioni,
// käsurealt või mingis muus keskkonnas, võib olla kausta failid asukoht
// relatiivse path'ina erinev. Loe täpsemalt kommentaari muutuja `FailidDir`
// juures.

void walk(const fs::path &path) {
  if (is_directory(path)) {
    std::cout << path.string() << "/\n";
    const auto it = fs::directory_iterator(path);
    for (const auto &entry : it) {
      walk(entry.path());
    }
  } else {
    auto content = readFileUntil(path, 10);
    std::cout << path.string() << " -> '" << content << "'\n";
  }
}

// 4. Ülesanne
// Loo `enum class`, mis võib olla 2 väärtust: `File` või `Dir`.
// Näide:
//   FileType::File
//   FileType::Dir
// mõlemad read peavad olema valiidsed C++ keeles.
// Lisaks vaata, et järjekord oleks õige (selle kohta ka kontroll paar rida
// allpool). Seejärel kirjuta funktsioon `fileType`, mis võtab argumendiks
// `fs::path` objekti ning tagastab, kas antud failitee viitab failile või
// kaustale. Vastavalt `FileType::File` ja `FileType::Dir`. Funktsiooni
// signatuuris ei tohi olla kirjas võtmesõna `auto`, s.t `auto fileType(...)` ei
// ole lubatud. Uuri, mis on vahet `enum` ja `enum class` konstruktsioonidel.
// Miks peaks eelistama üht üle teise? NB! Pane tähele, et viimasele küsimusele
// vastamine on ülesande läbimiseks kohustuslik, s.t on väärt punkte.

enum class FileType {
  File,
  Dir,
};

static_assert(static_cast<int>(FileType::File) == 0);
static_assert(static_cast<int>(FileType::Dir) == 1);

FileType fileType(const fs::path &path) {
  if (is_directory(path)) {
    return FileType::Dir;
  }
  return FileType::File;
}

// 5. Ülesanne
// Kirjuta funktsioon `writeFile`, mis võtab 3 argumenti: failitee (fs::path),
// sõne (std::string) ning lipu `shouldAppend` (bool), mille algväärtus on
// `false`. Funktsioon kirjutab antud sõne antud failiteel asuvasse faili.
// Funktsioon peab faili puudumisel selle looma.
// Juhul kui kolmas argument `shouldAppend` on `true`, ei tohi funktsioon
// failisisu üle kirjutada, vaid peab antud sõne lisama faili lõppu. Vastasel
// juhul peab funktsioon faili sisu üle kirjutama.
//
// Näide:
//   writeFile("a.txt", "tere") kirjutab faili "a.txt" sisu "tere", funktsioon
//   kirjutab sisu üle writeFile("a.txt", "tere", true) kirjutab faili "a.txt"
//   sisu "tere", funktsioon ei kirjuta sisu üle

void writeFile(const fs::path &path, const std::string &content,
               bool shouldAppend = false) {
  std::ios_base::openmode mode = std::ios::out;
  if (shouldAppend) {
    mode |= std::ios::app;
  }
  std::ofstream file(path, mode);
  file << content;
}

// [Lisaülesanne] 6. Ülesanne
// All on antud paar `static_assert` kontrolli. Kõik neist töötavad
// kompilaatoriga "Apple Clang 16.0" standardiga C++20. Implementeeri
// kompileerimisaegsed funktsioonid nii, et võimalikult palju neist
// `static_assert` testidest töötaksid korrektselt. Pane tähele, et mõni avaldis
// ei pruugi üldse teise kompilaatoriga olla lahendatav (s.t ei toeta
// kompileerimisaegseid avaldisi). Uuri, millised konstruktsioonid selliseid
// avaldisi toetavad. Vihje: võtmesõna `constexpr`. NB! `static_assert` kutseid
// muutma ei pea! Implementeeri antud funktsioon või jäta kommenteerituks.

constexpr long long int factorial(const long long int n) {
  if (n <= 1) {
    return n;
  }

  return n * factorial(n - 1);
}

constexpr std::array<int, 2> getArray() { return {1, 2}; }

constexpr std::vector<int> getVector() { return {1, 2}; }

constexpr std::string getString() { return "tere"; }

static_assert(factorial(1) == 1);
static_assert(factorial(2) == 2);
static_assert(factorial(3) == 6);

static_assert(std::is_same_v<decltype(getArray()), std::array<int, 2>>);
static_assert(getArray().size() == 2);
static_assert(getArray() == std::array{1, 2});

static_assert(std::is_same_v<decltype(getVector()), std::vector<int>>);
static_assert(getVector().size() == 2);

static_assert(std::is_same_v<decltype(getString()), std::string>);
static_assert(getString() == "tere");

// Ära neid muutujaid ja funktsioone redigeeri.
static int g_TotalTestCount = 0;
static int g_SuccessfulTestCount = 0;

#define TEST(condition) testAssert(condition, __FILE__, __LINE__)

void testAssert(bool condition, const char *file, int line) {
  g_TotalTestCount++;
  if (!condition) {
    std::cout << "VIGA: Kontroll asukohas " << file << ":" << line
              << " ei olnud õige.\n";
  } else {
    g_SuccessfulTestCount++;
  }
}

template <typename T> auto equal(const T &a, const T &b) {
  if (a.size() != b.size()) {
    return false;
  }

  return std::equal(a.begin(), a.end(), b.begin());
}

int MinuMain(const std::vector<std::string> &args) {
  // Igal programmil on alati vähemalt üks argument – tema ise. Uuri lähemalt
  // iseseisvalt.
  TEST(!args.empty());

  // Kontrollime, et oled kausta `failid` õigesse kohta pannud ja muutuja
  // `FailidDir` õigesti väärtustanud.
  TEST(fs::is_directory(FailidDir));

  // Allolev rida väljastab su töökausta ("working directory"):
  std::cout << "Programm jookseb kaustas " << fs::current_path() << '\n';

  // Ülesanne 1. Failide lugemine
  TEST(readFileUntil(FailidDir / "a.txt", 10).size() == 10);
  TEST(readFileUntil(FailidDir / "a.txt", 10) == "see on fai");
  TEST(readFileUntil(FailidDir / "c" / "f.txt", 10).size() == 6);
  TEST(readFileUntil(FailidDir / "c" / "f.txt", 10) == "lyhike");
  TEST(readFileUntil(FailidDir / "c" / "e", 10) == "69420");

  // Ülesanne 2. Faili lugemine ridade kaupa, `std::pair`
  {
    // ZIP-arhiivis kaasa antud fail peaks andma järgneva tulemuse
    const auto pairs = readPairs(FailidDir / "paarid.txt");
    const std::vector<StrPair> expectedPairs = {
        {"must", "valge"}, {"sinine", "punane"}, {"1", "2"}};
    TEST(equal(pairs, expectedPairs));
  }

  // Ülesanne 3. Failipuu kõndimine
  // Funktsioon peaks töötama korrektselt ZIP-arhiivis antud kaustaga `failid`.
  // Vaata näidet ülesande püstituse juures
  walk(FailidDir);

  // Ülesanne 4. `enum class`
  TEST(fileType(FailidDir) == FileType::Dir);
  TEST(fileType(FailidDir / "a.txt") == FileType::File);

  // Ülesanne 5. Faili kirjutamine
  {
    // Ära seda test-koodi modifitseeri.
    // Siin on pisut trikke, et Sulle mitte lahendus päris ette öelda

    auto const fName = fs::path("test.txt");
    copy_file(FailidDir / "a.txt", fName);

    auto fileSize = [fName] {
      return static_cast<long long int>(
          std::ifstream(fName, std::ifstream::ate | std::ifstream::binary)
              .tellg());
    };

    try {
      const auto origSize = fileSize();
      writeFile(fName, "tere", true);
      TEST(fileSize() > origSize);

      writeFile(fName, "tere");
      TEST(fileSize() == std::string{"tere"}.size());
    } catch (std::exception &) {
      std::cout << "Programm krahhis! Üritan parandada algset seisundit!\n";
    }

    fs::remove(fName);
  }

  std::cout << "Testide tulemus: " << g_SuccessfulTestCount << "/"
            << g_TotalTestCount << '\n';
  return g_SuccessfulTestCount != g_TotalTestCount;
}

template <typename T = std::string>
concept MinuMainKontroll = requires(std::vector<T> &a) {
  { MinuMain(a) } -> std::same_as<int>;
};
static_assert(MinuMainKontroll<>);

int main(const int argc, char **argv) {
  std::vector<std::string> args(argc);
  for (int i = 0; i < argc; i++) {
    args[i] = argv[i];
  }

  return MinuMain(args);
}
