#pragma once

#include <utility>

// 3. ülesanne. Implementeeri klassimall `MinuVektor<T>`, mis käitub vastavalt kirjeldusele. Enamjaolt peaks klassimall käituma
// sarnaselt `std::vector<T>` mallile.
//
// Vektor peab dünaamiliselt suurenema elementide lisamisel. Vali ise sobiv algoritm suurendamiseks.
// Naiivne lahendus puhvri täitumisel seda 2x suurendada on OK.
//
// Vektoril on 3 konstruktorit:
// 1. Vaikekonstruktor, mis allokeerib vektori siseselt puhvri suurusega 2.
// 2. Konstruktor, mille argumendiks on üks `size_t` arv, mis allokeerib puhvri antud suurusega.
// 3. Konstruktor, mille argumentideks on `size_t` ja `T`, mis allokeerib puhvri antud suurusega ja väärtustab
//    kõik puhvri elemendid antud tüübi T väärtusega.
// Vektoril on destruktor, mis vabastab allokeeritud puhvri.
//
// NB! Kõik konstruktorid peavad allokeerima "alloleva" puhvri.
//
// Vektoril on järgmised avalikud liikmefunktsioonid:
// - `void push_back(T)`, mis paneb antud väärtuse vektori lõppu.
// - `std::optional<T> get(size_t)`, mis tagastab "optional" elemendi asukohal, mis on antud funktsiooni
//    argumendina. Kui sellisel kohal elementi ei ole (nt üritatakse lugeda väärtust vektorist väljas), tagastada
//    tühi `std::optional` (`std::nullopt`).
// - `size_t size()`, mis tagastab vektoris olevate elementide arvu. Pane tähele, et ei ole korrektne tagastada
//    alloleva puhvri suurust.
// - `size_t capacity()`, mis tagastab puhvrisse mahtuvate elementide arvu.
// - `const T *buf() const`, mis tagastab alloleva puhvri. See funktsioon on testimiseks.
// - `std::optional<T> front()`, mis tagastab elemendi vektori alguses. Pane tähele, et see peab käituma
//    sarnaselt liikmefunktsioonile `get(size_t)` ehk tagastama tühja `std::optional` objekti juhul kui ei ole
//    võimalik esimest elementi tagastada. Mõtle, millal selline olukord on võimalik.
// - `std::optional<T> back()`, mis on ideelt sama nagu `front()` aga tagastab vektori viimase "optional" elemendi.
//
// Vektoril on järgmised avalikud operaatorid:
// - `T operator[](size_t)` mis tagastab väärtuse argumendi kohal. Sisuliselt on tegemist liikmefunktsiooniga `get`,
//    aga siin võid eeldada, et antud kohal väärtus eksisteerib.
// - Väljundoperaator, et saaks vektori väljastada järgmiselt: `std::cout << v << '\n';`.
//    Formaat peaks olema järgnev: "{x, y, z, ..., w}", kus elemendid x, y, z, ... ja w on vektori elemendid.
//    Vihje: Väljundoperaatori korrektne loomine käib pisut teistmoodi kui teiste operaatoride loomine. Uuri iseseisvalt,
//    kuidas seda teha.
//
// Lisaülesanne 1. (kuni 0.2 punkti)
// Loo klassile `MinuVektor<T>` staatiline funktsioon `iota(size_t)`, mis tagastab uue vektori antud pikkusega, mille
// elemendid on [0, 1, ..., n - 1], kus `n` on funktsiooni argument.
// Samuti peab funktsioonil olema piirang, et seda saab kutsuda vaid tüübiparameetritega, mis on täisarvutüüpi (0.1 punkti).
//
// Lisaülesanne 2. (kuni 0.3 punkti)
// Loo klassile `MinuVektor<T>` liikmefunktsioon `MinuVektor<U> transform(auto fn)` (antud argument on täpsemalt tüüpi
// `std::function<U(T &)> fn`, kasutame seal auto, et oleks kergem). Funktsioon konstrueerib uue vektori malliparameetriga
// `U` ning transformeerib iga käesoleva vektori elemendi antud funktsiooni alusel.
// Sisuliselt transformeerib ("mapib") funktsioon vektori elemendid antud funktsiooni alusel.

// NB! Pane tähele, et klassimalli pead tervenisti deklareerima ja defineerima (implementeerima) päisefailis.
// Lisaks on reeglina klassimalli liikmefunktsioone lihtsam implementeerida klassideklaratsiooni sees, mitte väljas.
// Vt näide failis `MinuPaar.hpp`

template<typename T>
class MinuVektor {
public:
    MinuVektor() {
        m_Buffer = new T[m_Capacity];
    }

    explicit MinuVektor(size_t capacity)
        : m_Capacity(capacity) {
        m_Buffer = new T[capacity];
    }

    MinuVektor(size_t capacity, T initial)
        : m_Capacity(capacity) {
        m_Buffer = new T[capacity];
        m_Size = capacity;
        for (size_t i = 0; i < m_Capacity; i++) {
            m_Buffer[i] = initial;
        }
    }

    ~MinuVektor() {
        delete[] m_Buffer;
    }

    void push_back(T value) {
        if (m_Size >= m_Capacity) {
            resize(m_Capacity * 2);
        }

        m_Buffer[m_Size++] = value;
    }

    std::optional<T> get(size_t idx) {
        if (idx >= m_Size) {
            return std::nullopt;
        }

        return m_Buffer[idx];
    }

    std::optional<T> front() {
        return get(0);
    }

    std::optional<T> back() {
        return get(m_Size - 1);
    }

    T operator[](size_t idx) {
        return *get(idx);
    }

    template<typename U>
    requires std::integral<U>
    static MinuVektor<U> iota(size_t sz) {
        MinuVektor v(sz);

        for (size_t i = 0; i < sz; i++) {
            v.push_back(i);
        }

        return v;
    }

    template<typename U>
    MinuVektor<U> transform(auto fn) const {
        MinuVektor<U> v(size());
        for (auto i = 0u; i < size(); i++) {
            v.push_back(fn(m_Buffer[i]));
        }
        return v;
    }

    [[nodiscard]] const T *buf() const { return m_Buffer; }
    [[nodiscard]] auto size() const { return m_Size; }
    [[nodiscard]] auto capacity() const { return m_Capacity; }

private:
    void resize(const size_t newCapacity) {
        if (newCapacity < m_Capacity) {
            return;
        }

        T *newBuffer = new T[newCapacity];
        for (size_t i = 0; i < m_Capacity; i++) {
            newBuffer[i] = m_Buffer[i];
        }
        delete[] m_Buffer;
        m_Buffer = newBuffer;
        m_Capacity = newCapacity;
    }

private:
    size_t m_Size{0};
    size_t m_Capacity{2};
    T *m_Buffer = nullptr;
};

template<typename T>
std::ostream &operator<<(std::ostream &os, MinuVektor<T> &v) {
    os << '{';
    for (size_t i = 0; i < v.size(); i++) {
        os << v[i];
        if (i != v.size() - 1) {
            os << ", ";
        }
    }
    os << '}';
    return os;
}
