#pragma once

// Pane tähele, et faililaiend `hpp` on reeglina viiteks sellele, et on tegemist
// C++ päisefailiga. Sageli kasutatakse siiani `.h` faililaiendit, mis ei ole tehniliselt
// vale, aga võib tekitada arusaamatusi nendest tulenevaid kompileerimisvigu.

// Samuti ei ole oluline, kuidas faile nimetad. C++ ei piira seda, et ühes failis peab olema üks klass
// või et failinimi peab olema klassi nimi jne. Nimeta oma failid alati nii, kuidas ülesandes nõutud või selle puudumisel
// kuidas ise soovid ja loogiline on. Küll aga ära pane kõike ühte faili :^)

#include <cstdlib>
#include <string>

// "Tavalise" klassi saame implementeerida nii:
// Deklaratsioon on päisefailis
// Definitsioon (implementatsioon) on vastavas lähtekoodifailis (vt String.cpp)

class String {
public:
    String();
    explicit String(const char *s);
    ~String();

    // Pane tähele, et klasside sees saab ka
    // kasutada funktsioonide üledefineerimist
    void append(const char *s);
    void append(char c);
    void append(const String &s);

    // Pane tähele, et kuna see liikmefunktsioon tagastab andmetest
    // std::string objekti, siis võib see olla `const`.
    [[nodiscard]] auto str() const -> std::string;

    // Siin on lihtsaks võrdlemiseks 2 võrdlusoperaatorit nii
    // sõneklassi endaga kui ka const char* "alumise" tüübiga
    bool operator==(const String &other) const;
    bool operator==(const char *other) const;

    // Üherealisi get-set liikmefunktsioone ei ole võib-olla isegi
    // mõtet cpp-faili panna. Reeglina need väga tihti ei muutu ja
    // seepärast võib need deklareerida ja defineerida otse
    // klassi deklaratsiooni sees
    [[nodiscard]] auto size() const { return m_Size; }
    [[nodiscard]] const char *data() const { return m_Buffer; }

private:
    // Reeglina me ei taha, et keegi väljastpoole sõne pikkust
    // ise muudaks. Seega on antud liikmefunktsioon privaatne.
    void resize(size_t newCapacity);

    // Pane tähele ka seda, et `private`, `public` ja `protected`
    // sektsioone võib defineerida läbisegi, nii kuidas vaja.
    // Sageli on hea näidata, et siin on uus "plokk" privaatseid
    // liikmeid, olgu need siis muutujad või funktsioonid.
    // Lõpuks väga vahet ei ole, kuidas klassi struktuuri siin
    // failis üles ehitad.
private:
    // Pane tähele muutujate järjekorda. Siin on väga oluline tähelepanek
    // see, et muutujad initsialiseeritakse selles järjekorras, milles
    // nad siin on defineeritud. S.t kõigepealt väärtustatakse m_Size = 0,
    // siis m_Capacity = 10 jne. See määrab ka initsialiseerimise nimekirjas
    // (initializer list) väärtustamise järjekorra. Uuri ja mängi sellega
    // ringi iseseisvalt.
    size_t m_Size{0};
    size_t m_Capacity{10};
    char *m_Buffer;
};
