#include <iostream>

#include "Memory.hpp"
#include "MinuPaar.hpp"
#include "MinuVektor.hpp"
#include "String.hpp"
#include "Tester.hpp"
#include "Utilities.hpp"

// Soovitav on lahendada ülesandeid testide järjekorras
// 100% arvestuse saamiseks on vaja ära lahendada 3 ülesannet

int main() {
  // Näidisülesannete testid
  // Need ei lähe lõpliku arvestuse testidesse
  {
    // Vaikekonstruktor
    const String s;
    TEST(s.size() == 0);

    // Operaator `bool operator==(const char *)`
    TEST(s == "");
  }

  {
    String s("hello world!");
    TEST(s.size() == 12);
    TEST(s == "hello world!");

    s.append("hello world!!!");
    TEST(s.size() == 26);
    TEST(s == "hello world!hello world!!!");
  }

  {
    String s;
    s.append("hello");
    s.append(' ');
    s.append("world");
    s.append('!');

    const auto str = s.str();
    TEST(str == "hello world!");
  }

  {
    const String s1("hello");
    const String s2("hello");
    const String s3("world");
    // Operaator `bool operator==(const String &)`
    TEST(s1 == s2);

    // Pane tähele, et vahel oskab kompilaator ise vastandoperaatori `bool
    // operator!=(const String &)` implementeerida
    TEST(s1 != s3);
    TEST(s2 != s3);
  }

  {
    const MinuPaar<int, int> paar{};
    TEST(paar.first == 0);
    TEST(paar.second == 0);

    const MinuPaar paar2{10, 20};
    TEST(paar2.first == 10);
    TEST(paar2.second == 20);
  }

  // Need on siin testide loenduri lähtestamiseks
  g_SuccessfulTestCount = 0;
  g_TotalTestCount = 0;

  // 1. ülesanne. Funktsioonimallid
  {
    // 1.1. `swap(*, *)`
    int a = 1;
    int b = 2;
    swap(&a, &b);
    TEST(a == 2);
    TEST(b == 1);
  }

  {
    // 1.2. `swap(&, &)`
    int a = 1;
    int b = 2;
    swap(a, b);
    TEST(a == 2);
    TEST(b == 1);
  }

  // 2. ülesanne
  {
    // Tüübipiirangute kontroll
    TEST(testGenerateType<int>());
    TEST(testGenerateType<float>());
    TEST(!testGenerateType<const char *>());

    auto buffer = generate<int>(0);
    TEST(buffer == nullptr);
    delete[] buffer;

    buffer = generate<int>(100);
    TEST(buffer != nullptr);
    delete[] buffer;
  }

  // 3. ülesanne. Klassimall `MinuVektor<T>`
  {
    const MinuVektor<int> v;
    TEST(v.size() == 0);
    TEST(v.capacity() == 2);
    TEST(v.buf() != nullptr);
  }

  {
    const MinuVektor<int> v(10);
    TEST(v.size() == 0);
    TEST(v.capacity() == 10);
    TEST(v.buf() != nullptr);
  }

  {
    const MinuVektor v(10, 1);
    TEST(v.size() == 10);
    TEST(v.capacity() == 10);

    const auto buf = v.buf();
    TEST(all(buf, buf + v.size(), 1));
  }

  {
    MinuVektor<int> v(10);
    v.push_back(1);
    TEST(v.size() == 1);
    v.push_back(2);
    TEST(v.size() == 2);

    MinuVektor<int> v2;
    for (int i = 0; i < 100; i++) {
      v2.push_back(i);
    }
    TEST(v2.size() == 100);
  }

  {
    MinuVektor<int> v(10);
    auto opt = v.get(0);
    TEST(opt == std::nullopt);

    v.push_back(1);
    opt = v.get(0);
    TEST(opt.value() == 1);

    opt = v.get(1);
    TEST(opt == std::nullopt);
  }

  {
    MinuVektor<int> v(10);
    auto opt = v.front();
    TEST(opt == std::nullopt);
    opt = v.back();
    TEST(opt == std::nullopt);
  }

  {
    MinuVektor<int> v(10);
    v.push_back(1);
    v.push_back(2);
    TEST(v[0] == 1);
    TEST(v[1] == 2);
  }

  {
    MinuVektor<int> v(10);
    for (int i = 0; i < 10; i++) {
      v.push_back(i);
    }

    std::stringstream buf;
    if (CoutCapture cap(buf); true) {
      std::cout << v;
    }
    TEST(buf.str() == "{0, 1, 2, 3, 4, 5, 6, 7, 8, 9}");
  }

  // Lisaülesanded
  // Testide lubamiseks muuda 0 -> 1
#if 0
    {
        // LÜ 1
        auto v = MinuVektor<int>::iota(10);
        bool result = true;
        for (size_t i = 0; i < 10; i++) {
            auto correct = v[i] == i;
            result &= correct;
        }
        TEST(result);
    }
#endif

#if 0
    {
        // LÜ 2
        auto v = MinuVektor<int>::iota(10);
        const auto fn = [](int i) { return static_cast<float>(i) + 0.25f; };
        auto fv = v.transform<float>(fn);

        bool result = true;
        for (size_t i = 0; i < 10; i++) {
            auto correct = fv[i] == static_cast<float>(i) + 0.25f;
            result &= correct;
        }
        TEST(result);
    }
#endif

  std::cout << "Testide tulemus: " << g_SuccessfulTestCount << "/"
            << g_TotalTestCount << '\n';
  return g_SuccessfulTestCount != g_TotalTestCount;
}
