#include "String.hpp"

#include <cassert>
#include <cstring>

// Vaikekonstruktor initsialiseerib lihtsalt puhvri
String::String() : m_Buffer(new char[m_Capacity]) {}

// Konstruktor `const char *` tüübist teeb
// sellest korrektselt sõne ning jätab puhvrisse
// pisut rohkem ruumi, et ei peaks kohe seda suurendama
// hakkama
// NB! Pane tähele liikmemuutujate initsialiseerimise
// järjekorda!
String::String(const char *s)
    : m_Size(strlen(s)), m_Capacity(m_Size * 2),
      m_Buffer(new char[m_Capacity]) {
  strncpy(m_Buffer, s, m_Size);
}

// Destruktor vabastab allokeeritud mälu
// Pane tähele, et on kasutatud `delete[]` ning
// võime alati eeldada, et `m_Buffer` on allokeeritud
// korrektselt. Miks?
String::~String() { delete[] m_Buffer; }

// Liikmefunktsioon `append(const char *)`
// Mõtle iseseisvalt, mida see funktsioon teeb ning
// kas see on korrektne. Kas siin võib olla probleeme?
void String::append(const char *s) {
  auto len = strlen(s);
  auto newLen = m_Size + len;
  if (newLen >= m_Capacity) {
    resize(newLen);
  }
  memcpy(m_Buffer + m_Size, s, len);
  m_Size += len;
  m_Buffer[m_Size] = '\0';
}

// Liikmefunktsioon `append(char)`
// Suurendame samuti puhvrit hetkel naiivselt 2x
void String::append(char c) {
  if (m_Size + 1 >= m_Capacity) {
    resize(m_Size * 2);
  }

  m_Buffer[m_Size] = c;
  m_Size += 1;
  m_Buffer[m_Size] = '\0';
}

// Liikmefunktsioon `append(const String &)`
// Saame ära kasutada liikmefunktsiooni `data`.
// Kas seda funktsiooni saaks teha optimaalsemaks?
// Näiteks nii, et igal kutsel ei pea funktsioonis
// `append(const char *)` sõne pikkust uuesti arvutama,
// kuna siin tehniliselt teame seda?
void String::append(const String &s) { append(s.data()); }

// Privaatne liikmefunktsioon `resize(size_t)`
// Suurendab puhvrit ning kopeerib eelmised elemendid ümber
void String::resize(size_t newCapacity) {
  assert(m_Buffer != nullptr);

  // Hetkel ei ole siin mõtet sõne väiksemaks teha, seega
  // selleks siin see assert.
  assert(newCapacity > m_Capacity);

  // Mõtle hoolega, mida see koodilõik siin teeb
  auto newBuffer = new char[newCapacity];
  memcpy(newBuffer, m_Buffer, m_Size);
  delete[] m_Buffer;
  m_Buffer = newBuffer;
  m_Capacity = newCapacity;
}

// Liikmefunktsioon `std::string str() const`
// Konstrueerime std::string objekti
// Pane tähele, et see funktsioon on `const`
std::string String::str() const { return std::string(data()); }

// Võrdlusoperaator `bool operator==(const String &other)`
// Pane tähele, et argument saab olla `const` kuna liikmefunktsioon
// `const char *data() const` on `const`, mistõttu saame seda `const`
// objektil kasutada. Lisaks saame välja kutsuda teise signatuuriga
// operaatori.
bool String::operator==(const String &other) const {
  return operator==(other.data());
}

bool String::operator==(const char *other) const {
  return strcmp(data(), other) == 0;
}
