#pragma once

#include <cstdlib>
#include <vector>

// Ülesanne 1. Funktsiooniobjektid ("funktor").
// Kirjuta klass `Sequence`, mis modelleerib sisult jada, mis hoiab sealt elementide võtmisel järge ning selle klassi
// objekti "kutsumisel" tagastatakse järjekorras järgmine element.
// Klass peab sisemiselt andmeid hoidma kuhimälus oleval mälul. Klass ei tohi kasutada STL-is olemasolevat konteinerit,
// vaid peab ise dünaamiliselt mälu haldama.
// Võid eeldada, et jadasse elemente peale selle konstrueerimist juurde ei lisata.
// Vihjeks: on vajalik vaid üks `new` ja `delete`. Puhvrit suuremaks ega väiksemaks tegema ei pea.
//
// Klass peab käituma järgmiselt:
//
// Klassil on 2 konstruktorit:
//  - `Sequence(const std::vector<int> &values)` - konstrueerib jada ning kopeerib kõik väärtused kuhimällu.
//  - `Sequence(const std::initializer_list<int> list)` - konstrueerib jada ning kopeerib initsialiseerimisnimekirja
//     elemendid kuhimällu.
// Klass ei tohi olla vaikimisi konstrueeritav! S.t, et `Sequence s;` ei tohi töötada (näites olemas).
// Klassi kustutamisel peab destruktor küsitud mälu vabastama.
//
// Klass peab olema "kutsutav" (funktsiooniobjekti omadustega), s.t, et sellele on "kutseoperaator".
//
// Klassil peavad olema järgmised kutseoperaatorid
//  - Ilma argumendita operaator, mis tagastab jada järjekorras järgmise elemendi ning liigutab "järge" edasi.
//  - Operaator ühe argumendiga tüüpi `size_t skip`, mis jätab vahele `skip` elementi ning tagastab vastava elemendi.
//    Samuti liigutab "järge" edasi vastavalt vahele jäetud elementide arvust.
//
//    Näide:
//      Sequence s = {1, 2, 3, 4, 5};
//      const auto x = s();  // x = 1
//      // s() tagastaks hetkel 2
//      const auto y = s(1); // y = 3
//      const auto z = s();  // z = 4
//      const auto w = s(1); // peaks krahhima programmi
//
//  - Operaator, mille abil on võimalik klassiobjekt teisendada tõeväärtuseks (`bool`).
//    Näide:
//      const Sequence s = {1};
//      const bool b = static_cast<bool>(s); // b = true
//      // const bool b2 = s; // pane tähele, et operaator peab olema `explicit`, et see rida ei töötaks.
//
// Klassil on ka järgmised liikmefunktsioonid
// NB! Pane tähele, et liikmefunktsioonid, mis klassi sisu ei muuda, peavad olema märgitud võtmesõnaga `const`.
// Lisaks oleks hea, kui funktsioonid, mis tagastavad midagi, võiksid olla märgitud atribuudiga `[[nodiscard]]`.
//
// - `int *end()` - tagastab viimasest elemendist järgmise. Tegelikult on see allokeeritud puhvrist väljas, aga hea
//    markeering selleks, et oleme jõudnud puhvri lõppu. Kõik STL-i konteinerid teevad seda sarnaselt (vt std::vector::end()).
// - `size_t available()` - tagastab alles olevate elementide arvu.
// - `std::vector<int> vec()` - tagastab alles olevad elemendid uues vektoris ning ei tohi jada järge edasi liigutada.
//
// Klassil on järgmised staatilised funktsioonid:
// - `Sequence seq(const Sequence &s)` - mis konstrueerib ja tagastab jadas `s` alles jäänud elementidest uue jada.
//    Vihje: Antud funktsioon on lahenduv 2 reaga ;)
//
// Soovitus: Alusta lahendamist all antud mallist.
//
// NB! Pane tähele, et ilmutatud ("explicit") teste automaatselt ei tehta. Need vaatan käsitsi.

class Sequence {
public:
    Sequence() = delete;

    explicit Sequence(const std::vector<int> &values);

    Sequence(std::initializer_list<int> list);

    // !!! Ära neid 2 modifitseeri
    Sequence(const Sequence &) = delete;
    Sequence &operator=(const Sequence &) = delete;
    // !!!

    ~Sequence();

    static Sequence sec(const Sequence &s);

    [[nodiscard]] auto end() const;

    [[nodiscard]] auto available() const -> size_t;

    int operator()();

    int operator()(size_t skip);

    explicit operator bool() const;

    [[nodiscard]] std::vector<int> vec() const;

private:
    int *m_Head;
    size_t m_Size;
    int *m_Ptr{};
};

