#pragma once

#include <print>
#include <functional>
#include <iostream>
#include <sstream>

// Ära neid muutujaid ja funktsioone redigeeri.
static inline int g_TotalTestCount = 0;
static inline int g_SuccessfulTestCount = 0;

#define TEST(condition) testAssert(condition, __FILE__, __LINE__)

inline void testAssert(bool condition, const char *file, int line) {
    g_TotalTestCount++;
    if (!condition) {
        std::cout << "VIGA: Kontroll asukohas " << file << ":" << line << " ei olnud õige.\n";
    } else {
        g_SuccessfulTestCount++;
    }
}

template<typename T>
auto equal(const T &a, const T &b) {
    if (a.size() != b.size()) {
        return false;
    }

    return std::equal(a.begin(), a.end(), b.begin());
}

auto all(const auto begin, const auto end, const auto val) {
    const auto fn = [&val](const auto v) -> bool { return v == val; };
    return std::all_of(begin, end, fn);
}

auto all(const auto &c, const auto val) {
    return all(std::begin(c), std::end(c), val);
}

class CoutCapture {
    std::stringstream &m_Buffer;
    std::streambuf *m_Prevcout;

public:
    explicit CoutCapture(std::stringstream &buf) : m_Buffer(buf) {
        m_Prevcout = std::cout.rdbuf(m_Buffer.rdbuf());
    }

    ~CoutCapture() {
        std::cout.rdbuf(m_Prevcout);
    }

    explicit operator bool() const {
        return m_Prevcout != std::cout.rdbuf();
    }

    static bool expect(const std::string &str, auto fn) {
        std::stringstream buf;
        if (const CoutCapture cap(buf); cap) {
            if constexpr (std::is_same_v<decltype(fn()), bool>) {
                auto res = fn();
                if (!res) {
                    return false;
                }
            } else {
                fn();
            }
        }

        return buf.str() == str;
    }
};