#include <algorithm>
#include <iostream>
#include <map>
#include <numeric>
#include <ranges>

#include "MinuVektor.hpp"

#include "Funktsioonid.hpp"
#include "Sequence.hpp"
#include "Tester.hpp"

int main() {
  // Ülesanne 1. Sequence
  {
    // Konstruktor `Sequence(const std::vector<int> &values)`
    const std::vector v = {1, 2, 3};
    const Sequence s(v);
    TEST(s.available() == v.size());
  }

  {
    // Konstruktor `Sequence(const std::initializer_list<int> list)`
    const Sequence s = {1, 2, 3, 4};
    TEST(s.available() == 4);
  }

  {
    // operator()
    Sequence s = {1, 2, 3, 4};
    std::vector<int> v(s.available());
    for (size_t i = 0; i < 4; ++i) {
      v[i] = s();
    }
    TEST(v.size() == 4);
    TEST(equal(v, {1, 2, 3, 4}));
    TEST(s.available() == 0);

    // s(); // - see peaks programmi krahhima
  }

  {
    // Konstruktor `Sequence(const Sequence &s)`
    Sequence s = {1, 2, 3};
    s();
    const auto s2 = Sequence::sec(s);
    TEST(equal(s2.vec(), {2, 3}));
  }

  {
    // operator(size_t)
    Sequence s = {1, 2, 3, 4, 5};
    s();
    TEST(s(1) == 3);
    TEST(s.available() > 0);
    TEST(s() == 4);
    TEST(s() == 5);
    TEST(s.available() == 0);

    // s(); // - see peaks programmi krahhima
  }

  {
    [[maybe_unused]] Sequence s = {1};
    // s(1); // - see peaks programmi krahhima
  }

  {
    // operator bool
    Sequence s = {1};
    TEST(static_cast<bool>(s));
    s();
    TEST(!static_cast<bool>(s));
  }

  {
    // vec
    const Sequence s = {1, 2};
    const auto vec = s.vec();
    TEST(equal(vec, {1, 2}));

    Sequence s2 = {1, 2};
    TEST(s2(1) == 2);
    const auto vec2 = s2.vec();
    TEST(vec2.empty());
  }

  {
    Sequence s = {1, 2, 3, 4, 5, 6};

    int current = 1;
    bool result = true;

    // Siin on operator bool kasulik:
    while (s && result) {
      const auto correct = (current++) == s();
      result &= correct;
    }
    TEST(result);
  }

  // Ülesanne 2. Funktsioonid
  // Kirjuta lahendus korrektselt päise- ja lähtekoodifailidesse nii, et
  // allolevad testid läbiksid Vaata, et lisaksid ka lähtekoodifaili faili
  // CMakeLists.txt ning impordiksid päisefaili õigesti NB! Pane tähele, et
  // funktsioonid ei ole mallid, s.t peavad olema jaotatud nii päise- kui ka
  // lähtekoodifailide vahel.
  // Antud ülesande lahendamisel ignoreeri CLion'i antud hoiatusi koopiate ja
  // reference'ide kohta.
  {
    std::vector v = {1, 2, 3};
    const std::vector v2 = vLisaCopy(v, 4);
    TEST(equal(v2, {1, 2, 3, 4}));
    TEST(equal(v, {1, 2, 3}));
  }

  {
    std::vector v = {1, 2, 3};
    const auto &v2 = vLisaRef(v, 4);

    // Funktsioon peab tagastama sama vektori, mis talle anti
    TEST(&v == &v2);

    TEST(equal(v, {1, 2, 3, 4}));
    TEST(equal(v2, {1, 2, 3, 4}));
  }

  {
    std::vector v = {1, 2, 3};
    const auto *v2 = vLisaPtr(&v, 4);

    // Funktsioon peab tagastama sama vektori, mis talle anti
    TEST(&v == v2);

    TEST(equal(v, {1, 2, 3, 4}));

    // Mõtle ja katseta, kas allolev rida teeb vektorist koopia?
    const auto &v2Ref = *v2;
    TEST(equal(v2Ref, {1, 2, 3, 4}));
  }

  // Ülesanne 3. Vektor viitadest
  // Kirjuta funktsioon `std::vector<int *> iota(int n)`, mis tagastab vektori,
  // kus on kuhimälus allokeeritud täisarvud, mis sisaldavad väärtuseid [0, ...,
  // n]. Võid eeldada, et `n > 0`. Kirjuta funktsioon `destroy(std::vector<int
  // *> &v)`, mis vabastab kõik täisarvud ning seab vektori viidad võrduma
  // väärtusega `nullptr`. Vasta ka küsimusele, miks argument peab olema viite
  // tüüpi ning miks ei tohi vektorit ennast "kustutada".
  {
    const int N = 10;
    auto vec = iota(N);
    bool result = true;
    for (int i = 0; i < N; ++i) {
      const auto ptr = vec[i];
      const auto correct = *ptr == (i + 1);
      result &= correct;
    }
    TEST(result);
    destroy(vec);
    TEST(all(vec, nullptr));
  }

  std::cout << "Testide tulemus: " << g_SuccessfulTestCount << "/"
            << g_TotalTestCount << '\n';
  return g_SuccessfulTestCount != g_TotalTestCount;
}
