#include "Sequence.hpp"

#include <cassert>

Sequence::Sequence(const std::vector<int> &values)
    : m_Size(values.size()) {
    m_Head = new int[m_Size];
    m_Ptr = m_Head;

    for (size_t i = 0; i < m_Size; ++i) {
        m_Ptr[i] = values[i];
    }
}

Sequence::Sequence(const std::initializer_list<int> list)
    : m_Size(list.size()) {
    m_Head = new int[m_Size];
    m_Ptr = m_Head;

    auto ptr = list.begin();
    for (size_t i = 0; i < m_Size; ++i, ++ptr) {
        m_Ptr[i] = *ptr;
    }
}

Sequence::~Sequence() {
    delete []m_Head;
}

auto Sequence::end() const {
    return m_Head + m_Size;
}

auto Sequence::available() const -> size_t {
    return end() - m_Ptr;
}

int Sequence::operator()() {
    assert(available());

    const auto value = *m_Ptr;
    m_Ptr++;
    return value;
}

int Sequence::operator()(const size_t skip) {
    assert(skip < available());

    for (size_t i = 0; i < skip; ++i) {
        m_Ptr++;
    }
    const auto value = *m_Ptr;
    m_Ptr++;

    return value;
}

Sequence::operator bool() const {
    return available() != 0;
}

std::vector<int> Sequence::vec() const {
    const auto size = available();
    std::vector<int> v(size);
    for (size_t i = 0; i < size; ++i) {
        v[i] = m_Ptr[i];
    }

    return v;
}

Sequence Sequence::sec(const Sequence &s) {
    const auto vec = s.vec();
    return Sequence(vec);
}
