#include <iostream>
#include <sstream>

// Proovi lahendada allolevad ülesanded. Lahendusi vaatame põgusalt järgmises praktikumis ja lahendused panen Git'i
// järgmise nädala alguses.

// 1. Ülesanne
// Kirjuta funktsioon, mis lühendab sõnena antud inimese nime antud näidete põhjal.
// Ära kasuta klassi `std::stringstream`
//
// Näide:
//   lyhenda("Mari-Anne Lill") = "Lill M-A."
//   lyhenda("Otto Triin Kaido Maastik") = "Maastik O. T. K."

// ...

// 2. Ülesanne
// Kirjuta funktsioon tagurpidi, mis keerab argumendina antud sõnevektori sõned tagurpidi.
// Seejärel keerab funktsioon vektori enda tagurpidi. Tagasta muudetud vektor.
//
// Näide:
//   tagurpidi({"auto", "koer"}) = {"reok", "otua"}

// ...

// 3. Ülesanne
// Kirjuta funktsioon, mis saab argumendiks sõnedevektori ning kustutab sealt kõik sõned, mis sisaldavad
// teises argumendis saadud sõne. S.t teine argument on kustutatava sõne alamsõne. Tagasta muudetud vektor.
//
// Näide:
//   kustuta({"tere", "auto", "koer"}, "er") = {"auto"}

// ...


// Võid allolevad kommentaarid eemaldada, et näha, millal mingit puhvrit suuremaks tehakse või uus objekt luuakse
// Reeglina ei ole see hea praktika, aga saame pisut rohkem informatsiooni allokatsioonide kohta.
// Allokatsioone ja võtmesõnu `new` ja `delete` vaatame tulevatel nädalatel.
//
// void *operator new(size_t size) {
//     std::println("new({})", size);
//     void *p = malloc(size);
//     return p;
// }
//
// void operator delete(void *p) {
//     std::println("delete()");
//     free(p);
// }

// Lahkame seda funktsiooni tulevate nädalatel
template<typename Container>
void prindi(Container &c) {
    std::cout << "[";
    for (auto x : c) {
        std::cout << x << ' ';
    }
    std::cout << "]\n";

    // Kui sul töötab std::println, kasuta seda
    // std::println("{:}", c);
}

int main() {
    {
        // Loob tühja vektori, kus pole ühtegi elementi
        // ega ruumi ühelegi elemendile
        std::vector<int> vec;

        // vec[0] = 1;
        // ^ Proovi seda rida jooksutada, see krahhib programmi
    }

    {
        // Loob tühja vektori, kus pole ühtegi telementi.
        // (Liikme)funktsioon reserveerib vektoris ruumi 10 elemendile, kuid ei initsialiseeri ühtegi elementi.
        std::vector<int> vec;
        vec.reserve(10);

        // vec[0] = 1;
        // ^ Kas siin on see rida OK?
    }

    {
        // Loob vektori, kus on ruumi 10 elemendile
        // ning initsialiseerib need tavaväärtusega

        std::vector<int> vec(10);

        // Loob vektori, aga elemendid on väärtustatakse väärtusega 1
        std::vector<int> vec1(10, 1);
    }

    {
        std::vector<int> vec(10);

        // Klassikaline for-tsükkel, kuidas elemente väärtustada
        for (size_t i = 0; i < vec.size(); i++) {
            vec[i] = 1;
        }

        // Miks allolev for-of-tsükkel ei väärtusta elemente?
        // Mis on tüüpi on muutuja `x`?
        for (auto x : vec) {
            x = 1;
        }

        // Aga miks allolev for-of-tsükkel töötab?
        for (auto &x : vec) {
            x = 1;
        }
    }

    {
        // Vahel on vektori konstrueerimisel võimalik ära jätta elementide tüübid
        // See on võimalik, kui konstrueerid vektori kohe elementidega
        std::vector vec{1, 2, 3};

        // või kasutad konstruktorit, mis määrab algväärtuse
        std::vector vec1(10, 0);

        // See on võimalik tänu CTAD feature'ile
        // Seda vaatame tulevikus põhjalikumalt, aga võid uurida lähemalt:
        // https://en.cppreference.com/w/cpp/language/class_template_argument_deduction
    }

    {
        // Miks selline vektori konstrueerimine on ebaefektiivne?
        // Mõtle, kuidas dünaamilist vektorit implementeerida
        // Paljud IDE-d peaksid tegema for-tsükli või seal sees midagi kollaseks ja viitama, et selline
        // koodilõik on ebaefektiivne ja soovitab paremini. Vaata, kas ja kuidas sinu IDE seda teeb.
        std::vector<int> vec;
        for (int i = 0; i < 10; i++) {
            vec.push_back(i);
        }
    }

    {
        // Mõtle, millal kasutada `std::vector` asemel `std::array`
        // Vihje: `std::array` ei saa lisada rohkem elemente, kui esialgselt teise "klassi"-argumendiga defineeritud.
        std::array<int, 4> arr{1, 2, 3, 4};

        // NB! `std::array` suurus peab olema koodis kirjas kompileerimise ajal.
        // Näiteks allolev koodilõik ei ole korrektne:
        // int n = 0;
        // std::cin >> n;
        // std::array<int, n> arr;
        // Kui on vaja kasutaja sisendit kasutada konteineri suurusena, siis peame kasutama dünaamilist konteinerit
    }

    {
        // Massiivi (`std::array`) saab ka konstrueerida ainult suurusega
        // Mis väärtusega on antud massiivi elemendid?
        std::array<int, 10> arr{};
    }

    {
        // Allolev muutuja C-tüüpi sõne, mis on tüübiga `const char*`
        // Selle sõne kasutamine on limiteeritud, uuri iseseisvalt, mis limitatsioonid sellel on.
        //   Kas saame sinna karaktereid lisada?
        //   Kas saame seal karaktereid asendada?
        const char *s = "tere maailm!";

        // C-tüüpi sõne on sisuliselt karakterite massiiv, mis pannakse reeglina programmi binaari sisse.
        // C-tüüpi sõne peab standardi järgi lõppema karakteriga 0 (NULL-byte, '\0'), mis signaliseerib sõne lõpu.
        // Funktsioonid, mille signatuuris on `const char*` eeldavad reeglina, et NULL-byte on olemas.
        // See tähendab, et on juhte, et NULL-byte pole vaja. Mõtle miks see võib olla kasulik.
        // Erandjuhul
        char s1[] = "tere maailm!";
        // on see sõne pinumälul, millest räägime lähemalt tulevate nädalatel.
        // Selles sõnes saame muuta eraldiseisvaid karaktereid, aga ei saa sõne suurust muuta.

        // Sõne pikkuse saame funktsiooniga strlen (samuti eeldab sõne lõpus NULL-byte'i). Kuidas muidu sõne lõppu saada?
    }

    {
        // C-tüüpi massiivid on sarnased C-tüüpi sõnedele, ainus erinevus on see, et neid ei panda reeglina* programmi
        // binaari ning ei lõppe NULL-byte'iga.
        // Miks ei ole NULL-byte kasutamine võimalik massiiviga, aga on võimalik C-tüüpi sõnega?
        // Vihje: ASCII tabel
        // NULL-byte'i puudumine tähendab küll seda, et alati C-tüüpi massiivi edasi andmisel funktsiooni vmt
        // peame kaasa andma ka selle pikkuse.
        // Miks see on halb? Kas see võib olla ka hea?

        // Konstrueerib massiivi, kus on 3 elementi
        int arr[] = {1, 2, 3};

        // Uuri iseseisvalt, mida teevad järgmised konstrueerimised
        int arr1[3] = {1, 2, 3};
        int arr2[3] = {};
        int arr3[3] = {0};

        // Miks see ei kompileeru?
        // int arr4[];
    }

    {
        // Aga kuidas ma leian sõnest mingi karakteri või alamsõne?

        // Kasutame siin `using` konstruktsiooni, siis ei pea kogu aeg
        // `std::string` välja kirjutama. Pane tähele, et liigne selle kasutamine
        // võib teha koodi loetavuse halvemaks.
        // Uuri seda konstruktsiooni lähemalt: https://en.cppreference.com/w/cpp/keyword/using
        using Str = std::string;

        Str s = "see on pikk sõne ja siin on palju sõnu";
        Str as = "tere";
        // char as = 's';
        if (const auto pos = s.find(as); pos == Str::npos) {
            std::println("ei leidnud!");
        } else {
            std::println("{} asub positsioonil {}, alles jääb '{}'", as, pos, s.substr(pos));
        }
    }

    {
        // Miks on allolev kood reeglina* halb?
        // Vihje: võib juhtuda, et teeme peaaegu iga += korral sõnest koopia (Javas samuti)
        // Huvi korral uuri konseptsiooni "short string"
        std::string result;
        for (int i = 0; i < 100; i++) {
            result += "a";
        }

        // Jah, saame tegelt luua sama sõne ka nii
        std::string result1(100, 'a');
        // aga sageli ülesanne seda ei võimalda

        // Seepärast on parem tsüklites sõnesid luues kasutada klassi `std::stringstream`
        std::stringstream ss;
        for (int i = 0; i < 100; i++) {
            ss << "a";
        }
        // ja saame konstrueerida sõne järgnevalt:
        auto str = ss.str();
        // Jällegi hea koht, kus kasutada võtmesõna `auto`

        // Klass `stringstream` võimaldab veel mitmeid muid asju, uuri iseseisvalt lähemalt.
        if (ss.eof()) {
            // ...
        }
    }

    return 0;
}
